@@ -1,9 +1,12 @@
 /* vim:ai ts=4 sw=4
  * tproxy.c:
  *
- * This is a transparent proxy server. It is started from inetd to service
- * HTTP requests are transparently accepted and passed to the WWW proxy
- * for handling.
+ * This is a transparent proxy server. It can be started from inetd to
+ * service HTTP requests, or run as a stand-alone daemon. HTTP requests
+ * are transparently accepted and passed to the WWW proxy cache for
+ * handling. The HTTP request line is modified into a form usable by
+ * the WWW proxy cache. The destination WWW site name is extracted from
+ * the HTTP headers to avoid DNS lookups.
  */
 
 #include <sys/time.h>
@@ -14,27 +17,31 @@
 #include <unistd.h>
 #include <syslog.h>
 #include <signal.h>
-#include <wait.h>
 #include <pwd.h>
-#include <grp.h>
-#include <getopt.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
+#include <errno.h>
 
-/*
- * Beta testing a hostname translation file. It might be better to use a DBM file,
- * will have to see how well it works out. I probably will want to add regular
- * expressions as well.
- */
-#ifdef TRANSLATE_PATH
-# define TRANSLATE_NAME(h)	translate_name(h)
-# ifndef MAX_TRANSLATIONS
-#  define MAX_TRANSLATIONS	100
-# endif
-#else
-# define TRANSLATE_NAME(h)	h
+#ifdef NEED_GETOPT_H
+#include <getopt.h>
+#endif
+
+#ifdef IPFILTER
+#include <fcntl.h>
+#include <sys/ioctl.h>
+#include <netinet/in_systm.h>
+#include <netinet/ip.h>
+#include <netinet/tcp.h>
+#include <net/if.h>
+#include <netinet/ip_compat.h>
+#include <netinet/ip_fil.h>
+#include <netinet/ip_nat.h>
+#endif
+
+#if defined(SIGCLD) && !defined(SIGCHLD)
+#define SIGCHLD SIGCLD
 #endif
 
 /*
@@ -57,18 +64,28 @@ static uid_t getuserid(char *user);
 static uid_t getgroupid(uid_t uid);
 static int bind_to_port(ipaddr_t bind_ip, short bind_port);
 static int connect_to_proxy(ipaddr_t ip, short port);
+static void lookup_hostname(struct sockaddr_in *addr, char *hostname,
+							int hostlen, int needed);
 static void server_main_loop(int sock, ipaddr_t server_ip, short server_port);
-static void dead_server(int sig);
-static void trans_proxy(int sock, struct sockaddr_in *addr, int len, ipaddr_t server_ip, short server_port);
-#ifdef TRANSLATE_PATH
-static const char *translate_name(const char *name);
-#endif
+static void trans_proxy(int sock, struct sockaddr_in *addr,
+						ipaddr_t server_ip, short server_port);
+static int	get_site_from_headers(int sock, char *hostname, int hostlen,
+								  char *headers, int header_max,
+								  int *header_len);
+static void alarm_signal(int sig);
 
 /*
  * Command line switches.
  */
 static int				fully_transparent = 0;
 
+#ifdef IPFILTER
+/*
+ * The /dev/ipnat device node.
+ */
+static int				natdev = -1;
+#endif
+
 /*
  * Main loop.
  */
@@ -156,6 +173,17 @@ int main(int argc, char **argv)
 		usage(argv[0], "No server port was specified (or it was invalid).");
 	}
 
+#ifdef IPFILTER
+	/*
+	 * Now is a good time to open /dev/ipnat, before giving up our uid/gid.
+	 */
+	if ((natdev = open(IPL_NAT, O_RDONLY)) < 0)
+	{
+		perror("open(" IPL_NAT ")");
+		exit(1);
+	}
+#endif
+
 	/*
 	 * See if we should run as a server.
 	 */
@@ -167,9 +195,14 @@ int main(int argc, char **argv)
 		sock = bind_to_port(bind_ip, bind_port);
 
 		/*
-		 * Start a server process.
+		 * Start a server process. When DEBUG is defined, just run
+		 * in the foreground.
 		 */
+#ifdef DEBUG
+		switch (0)
+#else
 		switch (fork())
+#endif
 		{
 		case -1:
 			perror("fork()");
@@ -185,23 +218,25 @@ int main(int argc, char **argv)
 			 * Ignore some signals.
 			 */
 			signal(SIGHUP, SIG_IGN);
+#ifndef DEBUG
 			signal(SIGINT, SIG_IGN);
+#endif
 			signal(SIGQUIT, SIG_IGN);
-			signal(SIGSTOP, SIG_IGN);
+			signal(SIGTSTP, SIG_IGN);
 			signal(SIGCONT, SIG_IGN);
 			signal(SIGPIPE, SIG_IGN);
+			signal(SIGALRM, alarm_signal);
 
 			/*
 			 * Drop back to an untrusted user.
 			 */
-			setregid(run_gid, run_gid);
-			setreuid(run_uid, run_uid);
+			setgid(run_gid);
+			setuid(run_uid);
 
 			/*
 			 * Start a new session and group.
 			 */
 			setsid();
-			setpgrp();
 
 			/*
 			 * Handle the server main loop.
@@ -237,10 +272,18 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
+	/*
+	 * Set the keepalive socket option to on.
+	 */
+	{
+		int		on = 1;
+		setsockopt(STDIN_FILENO, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof(on));
+	}
+
 	/*
 	 * We are running from inetd so process stdin.
 	 */
-	trans_proxy(STDIN_FILENO, &addr, len, server_ip, server_port);
+	trans_proxy(STDIN_FILENO, &addr, server_ip, server_port);
 	closelog();
 
 	return (0);
@@ -401,15 +444,16 @@ static int bind_to_port(ipaddr_t bind_ip, short bind_port)
 	 * Set the SO_REUSEADDR option for debugging.
 	 */
 	{
-	 	int	one = 1;
+	 	int	on = 1;
 
-		setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&one, sizeof(one));
+		setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
 	}
 #endif
 
 	/*
 	 * Set the address to listen to.
 	 */
+	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
 	addr.sin_port = htons(bind_port);
 	addr.sin_addr.s_addr = bind_ip;
@@ -443,25 +487,91 @@ static int connect_to_proxy(ipaddr_t ip, short port)
 	struct sockaddr_in	addr;
 	int					sock;
 
+	/*
+	 * Allocate a socket.
+	 */
 	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
 	{
 		syslog(LOG_WARNING, "socket(): %m");
 		return (-1);
 	}
 
+	/*
+	 * Set the address to connect to.
+	 */
+	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
 	addr.sin_port = htons(port);
 	addr.sin_addr.s_addr = ip;
 
+	/*
+	 * Connect our socket to the above address.
+	 */
 	if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)
 	{
 		syslog(LOG_WARNING, "connect(): %m");
 		return (-1);
 	}
 
+	/*
+	 * Set the keepalive socket option to on.
+	 */
+	{
+		int		on = 1;
+		setsockopt(STDIN_FILENO, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof(on));
+	}
 	return (sock);
 }
 
+/*
+ * Translate a sockaddr_in structure into a usable ASCII hostname.
+ */
+static void lookup_hostname(struct sockaddr_in *addr, char *hostname,
+							int hostlen, int needed)
+{
+	struct hostent	*host;
+
+	/*
+	 * Get the hostname IP in dotted decimal in case the lookup fails.
+	 */
+	strncpy(hostname, inet_ntoa(addr->sin_addr), hostlen);
+	hostname[hostlen - 1] = '\0';
+
+#ifdef DNS_LOOKUPS
+	/*
+	 * If the needed flag is set then always do a lookup.
+	 */
+	if (needed)
+	{
+		if ((host = gethostbyaddr((char *)&addr->sin_addr,
+								  sizeof(addr->sin_addr), AF_INET)) != NULL)
+		{
+			strncpy(hostname, host->h_name, hostlen);
+			hostname[hostlen - 1] = '\0';
+		}
+		else
+		{
+			syslog(LOG_INFO, "DNS lookup for %s failed: %m", hostname);
+		}
+	}
+#ifdef USELESS_DNS_LOOKUPS
+	else
+	{
+		if ((host = gethostbyaddr((char *)&addr->sin_addr,
+								  sizeof(addr->sin_addr), AF_INET)) != NULL)
+		{
+			strncpy(hostname, host->h_name, hostlen);
+			hostname[hostlen - 1] = '\0';
+		}
+		else
+		{
+			syslog(LOG_INFO, "DNS lookup for %s failed: %m", hostname);
+		}
+	}
+#endif
+#endif
+}
+
 /*
  * This is the main loop when running as a server.
  */
@@ -472,9 +582,20 @@ static void server_main_loop(int sock, ipaddr_t server_ip, short server_port)
 	int					len;
 
 	/*
-	 * Capture singals for dead servers.
+	 * Ignore dead servers so no zombies should be left hanging.
 	 */
-	signal(SIGCHLD, dead_server);
+#ifdef SA_NOCLDWAIT
+	{
+		struct sigaction	sa;
+
+		memset(&sa, 0, sizeof(sa));
+		sa.sa_handler = SIG_IGN;
+		sa.sa_flags = SA_NOCLDWAIT;
+		sigaction(SIGCHLD, &sa, NULL);
+	}
+#else
+ 	signal(SIGCHLD, SIG_IGN);
+#endif
 
 	for (;;)
 	{
@@ -484,11 +605,10 @@ static void server_main_loop(int sock, ipaddr_t server_ip, short server_port)
 		len = sizeof(addr);
 		while ((new_sock = accept(sock, (struct sockaddr *)&addr, &len)) < 0)
 		{
-			if (errno != EINTR)
-			{
-				syslog(LOG_ERR, "accept(): %m");
-				sleep(30);
-			}
+			/*
+			 * Connection resets are common enough to log them as debug only.
+			 */
+			syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_ERR), "accept(): %m");
 		}
 
 		/*
@@ -506,7 +626,7 @@ static void server_main_loop(int sock, ipaddr_t server_ip, short server_port)
 			/*
 			 * Start the proxy work in the new socket.
 			 */
-			trans_proxy(new_sock, &addr, len, server_ip, server_port);
+			trans_proxy(new_sock, &addr, server_ip, server_port);
 			close(new_sock);
 			closelog();
 			exit(0);
@@ -519,83 +639,70 @@ static void server_main_loop(int sock, ipaddr_t server_ip, short server_port)
 	}
 }
 
-/*
- * Catch dead servers and put them out of their misery.
- */
-static void dead_server(int sig)
-{
-	int		status;
-
-	/*
-	 * Restart the signal handler and wait for the dying server.
-	 */
-	signal(sig, dead_server);
-	wait(&status);
-}
-
 /*
  * Perform the transparent proxy activity.
  */
-static void trans_proxy(int sock, struct sockaddr_in *addr, int len, ipaddr_t server_ip, short server_port)
+static void trans_proxy(int sock, struct sockaddr_in *from_addr,
+						ipaddr_t server_ip, short server_port)
 {
 	struct sockaddr_in	to_addr;
 	int					to_len;
-	struct hostent		*host;
+	int					proxy;
 	char				from_host[64];
 	char				to_host[64];
-	int					read_len;
-	int					split;
+	static char			headers[16384 + 1];
+	int					header_len;
 	char				*uri_pos;
-	char				*nl;
-	int					proxy;
+	char				*uri_end;
 	int					max_fd;
 	fd_set				read_fd;
-	static char			line[256];
+	int					read_len;
+#ifdef IPFILTER
+	natlookup_t			natlook;
+#endif
 
 	/*
-	 * The first thing we do is get the IP address that was trying to be connected to.
+	 * The first thing we do is get the IP address that the client was
+	 * trying to connected to. Here lies part of the magic. Normally
+	 * getsockname returns our address, but not with transparent proxying.
 	 */
 	to_len = sizeof(to_addr);
 	if (getsockname(sock, (struct sockaddr *)&to_addr, &to_len) < 0)
 	{
-		syslog(LOG_WARNING, "getsockname(): %m");
+		syslog(LOG_ERR, "getsockname(): %m");
 		return;
 	}
 
+#ifdef IPFILTER
 	/*
-	 * Determine the hostnames and log a syslog entry.
+	 * However, under IPFILTER, we have to do more work. The "to_addr" has
+	 * been changed by the IP filter to trick the IP stack into processing
+	 * the packet locally. We need to lookup the real remote address
+	 * ourselves using an ioctl().
 	 */
-	strncpy(to_host, TRANSLATE_NAME(inet_ntoa(to_addr.sin_addr)), sizeof(to_host));
-#ifdef DNS_LOOKUPS
-	sethostent(0);
-	if ((host = gethostbyaddr((char *)&to_addr.sin_addr,
-								sizeof(to_addr.sin_addr), AF_INET)) != NULL)
-	{
-		strncpy(to_host, TRANSLATE_NAME(host->h_name), sizeof(to_host));
-	}
-	else
-	{
-		syslog(LOG_INFO, "DNS lookup for %s failed: %m", to_host);
-	}
-#endif
 
-	strncpy(from_host, TRANSLATE_NAME(inet_ntoa(addr->sin_addr)), sizeof(from_host));
-#ifdef DNS_LOOKUPS
-#ifdef USELESS_DNS_LOOKUPS
-	if ((host = gethostbyaddr((char *)&addr->sin_addr,
-								sizeof(addr->sin_addr), AF_INET)) != NULL)
-	{
-		strncpy(from_host, TRANSLATE_NAME(host->h_name), sizeof(from_host));
-	}
-	else
-	{
-		syslog(LOG_INFO, "DNS lookup for %s failed: %m", to_host);
-	}
-#endif
-	endhostent();
-#endif
+	/*
+	 * Build up the NAT natlookup structure.
+	 */
+	memset(&natlook, 0, sizeof(natlook));
+	natlook.nl_inip = to_addr.sin_addr;
+	natlook.nl_outip = from_addr->sin_addr;
+	natlook.nl_flags = IPN_TCP;
+	natlook.nl_inport = to_addr.sin_port;
+	natlook.nl_outport = from_addr->sin_port;
 
-	syslog(LOG_INFO, "Request from %s -> %s", from_host, to_host);
+	/*
+	 * Use the NAT device to lookup the mapping pair.
+	 */
+	if (ioctl(natdev, SIOCGNATL, &natlook) == -1)
+	{
+		syslog(LOG_ERR, "ioctl(SIOCGNATL): %m");
+		return;
+	}
+
+	to_addr.sin_addr = natlook.nl_realip;
+	to_addr.sin_port = natlook.nl_realport;
+#endif
 
 	/*
 	 * Connect to the proxy server.
@@ -603,30 +710,61 @@ static void trans_proxy(int sock, struct sockaddr_in *addr, int len, ipaddr_t se
 	proxy = connect_to_proxy(server_ip, server_port);
 
 	/*
-	 * If the daemon is operating in HTTP translation mode, then translate the
-	 * request line.
+	 * Fully transparent saves a lot of trouble, but then again it doesn't
+	 * support the transparent proxy feature. If the daemon is operating in
+	 * HTTP translation mode then translate the request line.
 	 */
-	if (!fully_transparent)
+	if (fully_transparent)
 	{
 		/*
-		 * Read the HTTP request line.
+		 * Lookup an ASCII representation of the host's IP address.
 		 */
-		switch (read_len = read(sock, line, sizeof(line)))
+		lookup_hostname(from_addr, from_host, sizeof(from_host), 0);
+		lookup_hostname(&to_addr, to_host, sizeof(to_host), 1);
+
+		/*
+		 * Log the facts about the connection.
+		 */
+		syslog(LOG_INFO, "Transparent %s -> %s", from_host, to_host);
+	}
+	else
+	{
+		/*
+		 * Lookup an ASCII representation of the from host's IP address.
+		 */
+		lookup_hostname(from_addr, from_host, sizeof(from_host), 0);
+
+		/*
+		 * See if the destination site can be found in the HTTP headers.
+		 */
+		if (get_site_from_headers(sock, to_host, sizeof(to_host),
+								  headers, sizeof(headers) - 1,
+								  &header_len))
 		{
-		case -1:
-			syslog(LOG_WARNING, "read() failed: %m");
-			/* Drop through */
-		case 0:
-			close(proxy);
-			return;
+			/*
+			 * Log the facts about the connection.
+			 */
+			syslog(LOG_INFO, "Request NoDNS %s -> %s", from_host, to_host);
+		}
+		else
+		{
+			/*
+			 * Fallback is to use the DNS to find the destination site.
+			 */
+			lookup_hostname(&to_addr, to_host, sizeof(to_host), 1);
+
+			/*
+			 * Log the facts about the connection.
+			 */
+			syslog(LOG_INFO, "Request %s -> %s", from_host, to_host);
 		}
 
 #ifdef PAY_THE_PENALTY
 		/*
-		 * Not that I'm mean or anything, but if they don't configure to use the
-		 * proxy directly then they pay a penalty.
+		 * Not that I'm mean or anything, but if they don't configure to
+		 * use the proxy directly then they pay a little penalty.
 		 */
-		sleep(5);
+		sleep(2);
 #endif
 
 		/*
@@ -635,64 +773,45 @@ static void trans_proxy(int sock, struct sockaddr_in *addr, int len, ipaddr_t se
 		 * with a / then we have transparently trapped a request, if it doesn't
 		 * then assume that we have been nominated as a proxy by the browser.
 		 */
-		for (uri_pos = NULL, split = 0; split < read_len; ++split)
-		{
-			/*
-			 * Stop when we find the space between the METHOD and /URI.
-			 */
-			if (line[split] == ' ')
-			{
-				uri_pos = &line[split];
-				break;
-			}
-		}
-
-		/*
-		 * Translate the URI if we found one and it's valid.
-		 */
-		if ((uri_pos != NULL) && (uri_pos[1] == '/'))
+		if (((uri_pos = strchr(headers, ' ')) != NULL) &&
+			(uri_pos[1] == '/'))
 		{
 			*uri_pos++ = '\0';
-			write(proxy, line, strlen(line));
+			write(proxy, headers, strlen(headers));
 			write(proxy, " http://", 8);
 			write(proxy, to_host, strlen(to_host));
-			write(proxy, uri_pos, read_len - strlen(line) - 1);
+			write(proxy, uri_pos, header_len - strlen(headers) - 1);
 
 			/*
 			 * Dump a nice log message showing what we have done.
 			 */
-			uri_pos[16] = '\0';
-			if (((nl = strchr(uri_pos, '\r')) != NULL)
-			    ||
-				((nl = strchr(uri_pos, '\n')) != NULL))
+			if ((uri_end = strchr(uri_pos, ' ')) != NULL)
 			{
-				*nl = '\0';
+				*uri_end = '\0';
+				syslog(LOG_DEBUG, "Translated '%s http://%s%s'", headers, to_host, uri_pos);
+#ifdef DEBUG
+				fprintf(stderr, "Translated '%s http://%s%s'\n", headers, to_host, uri_pos);
+#endif
 			}
-			else
-			{
-				strcat(uri_pos, "...");
-			}
-			syslog(LOG_DEBUG, "HTTP translated to '%s http://%s%s'", line, to_host, uri_pos);
 		}
 		else
 		{
-			write(proxy, line, read_len);
+			write(proxy, headers, header_len);
 
 			/*
 			 * Dump a nice log message showing what we have done.
 			 */
-			line[32] = '\0';
-			if (((nl = strchr(line, '\r')) != NULL)
-			    ||
-				((nl = strchr(line, '\n')) != NULL))
+			if ((uri_pos = strchr(headers, ' ')) != NULL)
 			{
-				*nl = '\0';
+				if ((uri_end = strchr(uri_pos + 1, ' ')) != NULL)
+				{
+					*uri_end = '\0';
+					syslog(LOG_DEBUG, "Untranslated '%s'", headers);
+#ifdef DEBUG
+					fprintf(stderr, "Untranslated '%s'\n", headers);
+#endif
+				}
 			}
-			else
-			{
-				strcat(line, "...");
-			}
-			syslog(LOG_DEBUG, "HTTP untranslated to '%s'", line);
 		}
 	}
 
@@ -709,34 +828,51 @@ static void trans_proxy(int sock, struct sockaddr_in *addr, int len, ipaddr_t se
 		FD_SET(proxy, &read_fd);
 		max_fd = FD_MAX(sock, proxy);
 
+		/*
+		 * Allow 5 minutes for any activity.
+		 */
+		alarm(300);
+
 		/*
 		 * Wait for some data to be read.
 		 */
 		if (select(max_fd + 1, &read_fd, NULL, NULL, NULL) < 0)
 		{
-			syslog(LOG_WARNING, "select(): %m");
+			syslog(LOG_ERR, "select(): %m");
 			close(proxy);
 			return;
 		}
 
+		/*
+		 * We have something so disable the alarm.
+		 */
+		alarm(0);
+
 		/*
 		 * See if any data can be read from the client.
 		 */
 		if (FD_ISSET(sock, &read_fd))
 		{
-			switch (read_len = read(sock, line, sizeof(line)))
+			switch (read_len = read(sock, headers, sizeof(headers) - 1))
 			{
 			case -1:
-				syslog(LOG_WARNING, "read(client) failed: %m");
+				syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_WARNING),
+						"read(client) failed: %m");
 				close(proxy);
 				return;
 
 			case 0:
+				close(sock);
 				close(proxy);
 				return;
 
 			default:
-				write(proxy, line, read_len);
+				if (write(proxy, headers, read_len) < 0)
+				{
+					syslog(LOG_WARNING, "write(proxy) failed: %m");
+					close(sock);
+					return;
+				}
 				break;
 			}
 		}
@@ -746,81 +882,115 @@ static void trans_proxy(int sock, struct sockaddr_in *addr, int len, ipaddr_t se
 		 */
 		if (FD_ISSET(proxy, &read_fd))
 		{
-			switch (read_len = read(proxy, line, sizeof(line)))
+			switch (read_len = read(proxy, headers, sizeof(headers) - 1))
 			{
 			case -1:
-				syslog(LOG_WARNING, "read(proxy) failed: %m");
-				close(proxy);
+				syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_WARNING),
+						"read(proxy) failed: %m");
+				close(sock);
 				return;
 
 			case 0:
 				close(proxy);
+				close(sock);
 				return;
 
 			default:
-				write(sock, line, read_len);
+				if (write(sock, headers, read_len) < 0)
+				{
+					syslog(LOG_WARNING, "write(client) failed: %m");
+					close(proxy);
+					return;
+				}
 				break;
 			}
 		}
 	}
 }
 
-#ifdef TRANSLATE_PATH
 /*
- * Use a table to translate hostname for which the reverse DNS lookup
- * is wrong or fails.
+ * Read in the headers for the HTTP reset and store in the headers array.
+ * Then search the headers for a Host: hostname[:port] header.
  */
-static const char *translate_name(const char *name)
+static int	get_site_from_headers(int sock, char *hostname, int hostlen,
+								  char *headers, int header_max,
+								  int *header_len)
 {
-	static char	*translate_table[MAX_TRANSLATIONS];
-	int			index;
-	static FILE	*fp = NULL;
-	char		line[128];
-	char		*new_name;
-	char		*nl;
+	int		read_len;
+	char	*start;
+	char	*end;
 
 	/*
-	 * Load the translation table if it's not already loaded.
+	 * Allow 2 minutes to read in the headers.
 	 */
-	if (fp == NULL)
+	alarm(120);
+
+	/*
+	 * Read in a buffers worth of data, hopefully all of the headers
+	 * will be contained in this initial read. This is tagged with a
+	 * big !!FIX ME!!
+	 */
+	switch (read_len = read(sock, headers, header_max))
+	{
+	case -1:
+		syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_WARNING),
+				"read(client) failed: %m");
+		return (-1);
+
+	case 0:
+		return (0);
+	}
+	*header_len = read_len;
+
+	/*
+	 * Terminate the buffer so the string functions don't run
+	 * of the end.
+	 */
+	headers[read_len] = '\0';
+
+	/*
+	 * See if the specified string exists in the buffer.
+	 */
+	if ((start = strstr(headers, "\r\nHost: ")) != NULL)
 	{
 		/*
-		 * Check if the file exists.
+		 * We found it, so now skip over it.
 		 */
-		if ((fp = fopen(TRANSLATE_PATH, "r")) != NULL)
+		start += strlen("\r\nHost: ");
+
+		/*
+		 * Now see if the field we want is terminated by one
+		 * the termination characters.
+		 */
+		if ((end = strpbrk(start, ":\r\n\t ")) != NULL)
 		{
 			/*
-			 * Load each line into the translate table.
+			 * We found the end, now copy the field to the output.
 			 */
-			index = 0;
-			while (fgets(line, sizeof(line), fp) != NULL)
+			while ((start < end) && (hostlen-- > 1))
 			{
-				if ((new_name = strchr(line, ':')) != NULL)
-				{
-					*new_name++ = '\0';
-					if ((nl = strchr(new_name, '\n')) != NULL)
-					{
-						*nl = '\0';
-						translate_table[index++] = strdup(line);
-						translate_table[index++] = strdup(new_name);
-					}
-				}
+				*hostname++ = *start++;
 			}
-			translate_table[index] = NULL;
-			fclose(fp);
+			*hostname = '\0';
+
+			/*
+			 * Return success.
+			 */
+			return (1);
 		}
 	}
 
-	for (index = 0; translate_table[index] != NULL; ++index)
-	{
-		if ((strcmp(name, translate_table[index]) == 0)
-			&&
-			(translate_table[index + 1] != NULL))
-		{
-			return (translate_table[index + 1]);
-		}
-	}
+	/*
+	 * Return failure.
+	 */
+	return (0);
+}
 
-	return (name);
+/*
+ * Catch alarm signals and exit.
+ */
+static void alarm_signal(int sig)
+{
+	syslog(LOG_DEBUG, "Alarm signal caught - connection timeout");
+	exit(1);
 }
-#endif
